@@ -1,6 +1,7 @@
 /* vi:set et ai sw=2 sts=2 ts=2: */
 /*-
  * Copyright (c) 2009-2011 Jannis Pohlmann <jannis@xfce.org>
+ * Copyright (c) 2015      Ali Abdallah    <ali@xfce.org>
  *
  * This program is free software; you can redistribute it and/or 
  * modify it under the terms of the GNU General Public License as
@@ -36,18 +37,13 @@
 
 #include <gio/gio.h>
 
-#include <dbus/dbus.h>
-#include <dbus/dbus-glib.h>
-#include <dbus/dbus-glib-lowlevel.h>
-
 #include <tumblerd/tumbler-component.h>
 #include <tumblerd/tumbler-manager.h>
-#include <tumblerd/tumbler-manager-dbus-bindings.h>
+#include <tumblerd/tumbler-manager-gdbus.h>
 #include <tumblerd/tumbler-specialized-thumbnailer.h>
 #include <tumblerd/tumbler-utils.h>
 
 
-
 /* Property identifiers */
 enum
 {
@@ -73,6 +69,11 @@ static void             tumbler_manager_set_property      (GObject          *obj
                                                            guint             prop_id,
                                                            const GValue     *value,
                                                            GParamSpec       *pspec);
+static gboolean         tumbler_manager_register_cb       (TumblerExportedManager *skeleton,
+                                                           GDBusMethodInvocation  *invocation,
+                                                           const gchar *const     *uri_schemes, 
+                                                           const gchar *const     *mime_types,
+                                                           TumblerManager         *manager);
 static void             tumbler_manager_monitor_unref     (GFileMonitor     *monitor,
                                                            TumblerManager   *manager);
 static void             tumbler_manager_load_thumbnailers (TumblerManager   *manager,
@@ -105,9 +106,12 @@ struct _TumblerManagerClass
 struct _TumblerManager
 {
   TumblerComponent __parent__;
-
-  DBusGConnection *connection;
-  TumblerRegistry *registry;
+  GDBusConnection        *connection;
+  TumblerExportedManager *skeleton;
+  
+  gboolean                dbus_interface_exported;
+  
+  TumblerRegistry        *registry;
 
   /* Directory and monitor objects for the thumbnailer dirs in 
    * XDG_DATA_HOME and XDG_DATA_DIRS */
@@ -157,7 +161,7 @@ static void
 tumbler_manager_class_init (TumblerManagerClass *klass)
 {
   GObjectClass *gobject_class;
-
+  
   gobject_class = G_OBJECT_CLASS (klass);
   gobject_class->constructed = tumbler_manager_constructed;
   gobject_class->finalize = tumbler_manager_finalize; 
@@ -165,11 +169,12 @@ tumbler_manager_class_init (TumblerManagerClass *klass)
   gobject_class->set_property = tumbler_manager_set_property;
 
   g_object_class_install_property (gobject_class, PROP_CONNECTION,
-                                   g_param_spec_pointer ("connection",
-                                                         "connection",
-                                                         "connection",
-                                                         G_PARAM_READWRITE |
-                                                         G_PARAM_CONSTRUCT_ONLY));
+                                   g_param_spec_object ("connection",
+                                                        "connection",
+                                                        "connection",
+                                                        G_TYPE_DBUS_CONNECTION,
+                                                        G_PARAM_READWRITE |
+                                                        G_PARAM_CONSTRUCT_ONLY));
 
   g_object_class_install_property (gobject_class, PROP_REGISTRY,
                                    g_param_spec_object ("registry",
@@ -205,20 +210,32 @@ tumbler_manager_init (TumblerManager *manager)
 static void
 tumbler_manager_constructed (GObject *object)
 {
-  TumblerManager *manager = TUMBLER_MANAGER (object);
+  TumblerManager *manager;
+  GError *error = NULL;
+  
+  manager = TUMBLER_MANAGER (object);
 
-  /* everything's fine, install the manager type D-Bus info */
-  dbus_g_object_type_install_info (G_OBJECT_TYPE (manager), 
-                                   &dbus_glib_tumbler_manager_object_info);
+  manager->skeleton = tumbler_exported_manager_skeleton_new ();
 
-  /* register the manager instance as a handler of the manager interface */
-  dbus_g_connection_register_g_object (manager->connection, 
-                                       "/org/freedesktop/thumbnails/Manager1", 
-                                       G_OBJECT (manager));
+  g_dbus_interface_skeleton_export (G_DBUS_INTERFACE_SKELETON(manager->skeleton),
+				                            manager->connection,
+				                            "/org/freedesktop/thumbnails/Manager1",
+				                            &error);
+	
+  if (error != NULL)
+    {
+      g_critical ("error exporting thumbnail manager on session bus: %s", error->message);
+      g_error_free (error);
+      manager->dbus_interface_exported = FALSE;
+    }
+  else
+    {
+      manager->dbus_interface_exported = TRUE;
+      g_signal_connect(manager->skeleton, "handle-register", 
+                       G_CALLBACK(tumbler_manager_register_cb), manager);
+    }
 }
 
-
-
 static void
 tumbler_manager_finalize (GObject *object)
 {
@@ -241,9 +258,20 @@ tumbler_manager_finalize (GObject *object)
   /* release the registry */
   g_object_unref (manager->registry);
 
-  /* release the D-Bus connection object */
-  dbus_g_connection_unref (manager->connection);
+  /* Unexport from dbus */
+  if (manager->dbus_interface_exported)
+    g_dbus_interface_skeleton_unexport_from_connection 
+      (
+        G_DBUS_INTERFACE_SKELETON (manager->skeleton),
+        manager->connection
+      );
 
+  /* release the Skeleton object */
+  g_object_unref (manager->skeleton);
+  
+  /* release the D-Bus connection object */
+  g_object_unref (manager->connection);
+  
   tumbler_mutex_unlock (manager->mutex);
 
   /* destroy the mutex */
@@ -265,7 +293,7 @@ tumbler_manager_get_property (GObject    *object,
   switch (prop_id)
     {
     case PROP_CONNECTION:
-      g_value_set_pointer (value, manager->connection);
+      g_value_set_object (value, manager->connection);
       break;
     case PROP_REGISTRY:
       g_value_set_object (value, manager->registry);
@@ -289,7 +317,7 @@ tumbler_manager_set_property (GObject      *object,
   switch (prop_id)
     {
     case PROP_CONNECTION:
-      manager->connection = dbus_g_connection_ref (g_value_get_pointer (value));
+      manager->connection = g_object_ref (g_value_get_object (value));
       break;
     case PROP_REGISTRY:
       manager->registry = g_value_dup_object (value);
@@ -300,7 +328,38 @@ tumbler_manager_set_property (GObject      *object,
     }
 }
  
- 
+static gboolean
+tumbler_manager_register_cb (TumblerExportedManager *skeleton,
+                             GDBusMethodInvocation  *invocation,
+                             const gchar *const     *uri_schemes, 
+                             const gchar *const     *mime_types,
+                             TumblerManager         *manager)
+{
+  TumblerThumbnailer *thumbnailer;
+  const gchar        *sender_name;
+
+  g_return_val_if_fail (TUMBLER_IS_MANAGER (manager), FALSE);
+  g_return_val_if_fail (uri_schemes != NULL, FALSE);
+  g_return_val_if_fail (mime_types != NULL, FALSE);
+  
+  sender_name = g_dbus_method_invocation_get_sender (invocation);
+
+  tumbler_mutex_lock (manager->mutex);
+
+  thumbnailer = tumbler_specialized_thumbnailer_new_foreign (manager->connection,
+                                                             sender_name, uri_schemes, 
+                                                             mime_types);
+
+  tumbler_registry_add (manager->registry, thumbnailer);
+
+  g_object_unref (thumbnailer);
+
+  tumbler_mutex_unlock (manager->mutex);
+
+  tumbler_exported_manager_complete_register(skeleton, invocation);
+  
+  return TRUE;
+}
  
 static void
 tumbler_manager_monitor_unref (GFileMonitor   *monitor,
@@ -1176,10 +1235,9 @@ tumbler_manager_load_thumbnailers (TumblerManager *manager,
  * directory objects, directory monitors, override infos and thumbnailer 
  * infos. 
  */
-static void
+void
 tumbler_manager_load (TumblerManager *manager)
 {
-  GHashTable         *single_path;
   const gchar *const *data_dirs;
   GFileMonitor       *monitor;
   GList              *directories = NULL;
@@ -1205,36 +1263,14 @@ tumbler_manager_load (TumblerManager *manager)
   /* determine system data dirs */
   data_dirs = g_get_system_data_dirs ();
 
-  /* Create a ghash table to insert loaded directory path to avoid duplication */
-  single_path = g_hash_table_new_full (g_file_hash, (GEqualFunc)g_file_equal, g_object_unref, NULL);
-
   /* build $XDG_DATA_DIRS/thumbnailers dirnames and prepend them to the list */
   for (n = 0; data_dirs[n] != NULL; ++n)
     {
-      GFile *path;
-
-      path = g_file_new_for_path(data_dirs[n]);
-
-      if (!g_hash_table_lookup (single_path, path))
-        {
-          /* Save it in the hash table so we can relocate it */
-          /* path will be free automatically by g_hash_table_destroy */
-          g_hash_table_insert (single_path, path, path);
-
-          dirname = g_build_filename (data_dirs[n], "thumbnailers", NULL);
-          directories = g_list_prepend (directories, g_file_new_for_path (dirname));
-          g_free (dirname);
-        }
-      else
-        {
-          /* Free the path GFile object */
-          g_object_unref(path);
-        }
+      dirname = g_build_filename (data_dirs[n], "thumbnailers", NULL);
+      directories = g_list_prepend (directories, g_file_new_for_path (dirname));
+      g_free (dirname);
     }
 
-  /* destroy the hash table used for loading single pathes */
-  g_hash_table_destroy (single_path);
-
   /* reverse the directory list so that the directories with highest 
    * priority come first */
   directories = g_list_reverse (directories);
@@ -1850,7 +1886,7 @@ dump_thumbnailers (TumblerManager *manager)
 
 
 TumblerManager *
-tumbler_manager_new (DBusGConnection         *connection,
+tumbler_manager_new (GDBusConnection         *connection,
                      TumblerLifecycleManager *lifecycle_manager,
                      TumblerRegistry         *registry)
 {
@@ -1861,81 +1897,9 @@ tumbler_manager_new (DBusGConnection         *connection,
                        NULL);
 }
 
-
-
-gboolean
-tumbler_manager_start (TumblerManager *manager,
-                       GError        **error)
+gboolean tumbler_manager_is_exported (TumblerManager *manager) 
 {
-  DBusConnection *connection;
-  gint            result;
-
   g_return_val_if_fail (TUMBLER_IS_MANAGER (manager), FALSE);
-  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
-
-  tumbler_mutex_lock (manager->mutex);
-
-  /* get the native D-Bus connection */
-  connection = dbus_g_connection_get_connection (manager->connection);
-
-  /* request ownership for the manager interface */
-  result = dbus_bus_request_name (connection, "org.freedesktop.thumbnails.Manager1",
-                                  DBUS_NAME_FLAG_DO_NOT_QUEUE, NULL);
-
-  /* check if that failed */
-  if (result != DBUS_REQUEST_NAME_REPLY_PRIMARY_OWNER)
-    {
-      if (error != NULL)
-        {
-          g_set_error (error, DBUS_GERROR, DBUS_GERROR_FAILED,
-                       _("Another thumbnailer manager is already running"));
-        }
-
-      tumbler_mutex_unlock (manager->mutex);
-
-      /* i can't work like this! */
-      return FALSE;
-    }
-
-  tumbler_mutex_unlock (manager->mutex);
-
-  /* load thumbnailers installed into the system permanently */
-  tumbler_manager_load (manager);
-
-  /* done initializing and loading all permanent thumbnailers */
-  return TRUE;
-}
-
-
-
-void
-tumbler_manager_register (TumblerManager        *manager, 
-                          const gchar *const    *uri_schemes, 
-                          const gchar *const    *mime_types, 
-                          DBusGMethodInvocation *context)
-{
-  TumblerThumbnailer *thumbnailer;
-  gchar              *sender_name;
-
-  dbus_async_return_if_fail (TUMBLER_IS_MANAGER (manager), context);
-  dbus_async_return_if_fail (uri_schemes != NULL, context);
-  dbus_async_return_if_fail (mime_types != NULL, context);
-
-  sender_name = dbus_g_method_get_sender (context);
-
-  tumbler_mutex_lock (manager->mutex);
-
-  thumbnailer = tumbler_specialized_thumbnailer_new_foreign (manager->connection,
-                                                             sender_name, uri_schemes, 
-                                                             mime_types);
-
-  tumbler_registry_add (manager->registry, thumbnailer);
-
-  g_object_unref (thumbnailer);
-
-  tumbler_mutex_unlock (manager->mutex);
-
-  g_free (sender_name);
-
-  dbus_g_method_return (context);
+ 
+  return manager->dbus_interface_exported;
 }
